@@ -2,7 +2,7 @@
 
   struct.c -
 
-  $Author: glass $
+  $Author$
   created at: Tue Mar 22 18:44:30 JST 1995
 
   Copyright (C) 1993-2007 Yukihiro Matsumoto
@@ -12,6 +12,7 @@
 #include "internal.h"
 #include "vm_core.h"
 #include "id.h"
+#include "transient_heap.h"
 
 /* only for struct[:field] access */
 enum {
@@ -23,7 +24,7 @@ const rb_iseq_t *rb_method_for_self_aref(VALUE name, VALUE arg, rb_insn_func_t f
 const rb_iseq_t *rb_method_for_self_aset(VALUE name, VALUE arg, rb_insn_func_t func);
 
 VALUE rb_cStruct;
-static ID id_members, id_back_members;
+static ID id_members, id_back_members, id_keyword_init;
 
 static VALUE struct_alloc(VALUE);
 
@@ -47,6 +48,12 @@ struct_ivar_get(VALUE c, ID id)
     }
 }
 
+VALUE
+rb_struct_s_keyword_init(VALUE klass)
+{
+    return struct_ivar_get(klass, id_keyword_init);
+}
+
 VALUE
 rb_struct_s_members(VALUE klass)
 {
@@ -132,7 +139,6 @@ static inline int
 struct_member_pos(VALUE s, VALUE name)
 {
     VALUE back = struct_ivar_get(rb_obj_class(s), id_back_members);
-    VALUE const * p;
     long j, mask;
 
     if (UNLIKELY(NIL_P(back))) {
@@ -142,7 +148,6 @@ struct_member_pos(VALUE s, VALUE name)
 	rb_raise(rb_eTypeError, "corrupted struct");
     }
 
-    p = RARRAY_CONST_PTR(back);
     mask = RARRAY_LEN(back);
 
     if (mask <= AREF_HASH_THRESHOLD) {
@@ -152,7 +157,7 @@ struct_member_pos(VALUE s, VALUE name)
 		     mask, RSTRUCT_LEN(s));
 	}
 	for (j = 0; j < mask; j++) {
-	    if (p[j] == name)
+            if (RARRAY_AREF(back, j) == name)
 		return (int)j;
 	}
 	return -1;
@@ -167,9 +172,10 @@ struct_member_pos(VALUE s, VALUE name)
     j = struct_member_pos_ideal(name, mask);
 
     for (;;) {
-	if (p[j] == name)
-	    return FIX2INT(p[j + 1]);
-	if (!RTEST(p[j])) {
+        VALUE e = RARRAY_AREF(back, j);
+        if (e == name)
+            return FIX2INT(RARRAY_AREF(back, j + 1));
+        if (!RTEST(e)) {
 	    return -1;
 	}
 	j = struct_member_pos_probe(j, mask);
@@ -211,7 +217,7 @@ rb_struct_getmember(VALUE obj, ID id)
     }
     rb_name_err_raise("`%1$s' is not a struct member", obj, ID2SYM(id));
 
-    UNREACHABLE;
+    UNREACHABLE_RETURN(Qnil);
 }
 
 static VALUE rb_struct_ref0(VALUE obj) {return RSTRUCT_GET(obj, 0);}
@@ -279,7 +285,7 @@ new_struct(VALUE name, VALUE super)
 static void
 define_aref_method(VALUE nstr, VALUE name, VALUE off)
 {
-    rb_control_frame_t *FUNC_FASTCALL(rb_vm_opt_struct_aref)(rb_thread_t *, rb_control_frame_t *);
+    rb_control_frame_t *FUNC_FASTCALL(rb_vm_opt_struct_aref)(rb_execution_context_t *, rb_control_frame_t *);
     const rb_iseq_t *iseq = rb_method_for_self_aref(name, off, rb_vm_opt_struct_aref);
 
     rb_add_method_iseq(nstr, SYM2ID(name), iseq, NULL, METHOD_VISI_PUBLIC);
@@ -288,16 +294,25 @@ define_aref_method(VALUE nstr, VALUE name, VALUE off)
 static void
 define_aset_method(VALUE nstr, VALUE name, VALUE off)
 {
-    rb_control_frame_t *FUNC_FASTCALL(rb_vm_opt_struct_aset)(rb_thread_t *, rb_control_frame_t *);
+    rb_control_frame_t *FUNC_FASTCALL(rb_vm_opt_struct_aset)(rb_execution_context_t *, rb_control_frame_t *);
     const rb_iseq_t *iseq = rb_method_for_self_aset(name, off, rb_vm_opt_struct_aset);
 
     rb_add_method_iseq(nstr, SYM2ID(name), iseq, NULL, METHOD_VISI_PUBLIC);
 }
 
+static VALUE
+rb_struct_s_inspect(VALUE klass)
+{
+    VALUE inspect = rb_class_name(klass);
+    if (RTEST(rb_struct_s_keyword_init(klass))) {
+	rb_str_cat_cstr(inspect, "(keyword_init: true)");
+    }
+    return inspect;
+}
+
 static VALUE
 setup_struct(VALUE nstr, VALUE members)
 {
-    const VALUE *ptr_members;
     long i, len;
 
     members = struct_set_members(nstr, members);
@@ -306,17 +321,18 @@ setup_struct(VALUE nstr, VALUE members)
     rb_define_singleton_method(nstr, "new", rb_class_new_instance, -1);
     rb_define_singleton_method(nstr, "[]", rb_class_new_instance, -1);
     rb_define_singleton_method(nstr, "members", rb_struct_s_members_m, 0);
-    ptr_members = RARRAY_CONST_PTR(members);
+    rb_define_singleton_method(nstr, "inspect", rb_struct_s_inspect, 0);
     len = RARRAY_LEN(members);
     for (i=0; i< len; i++) {
-	ID id = SYM2ID(ptr_members[i]);
+        VALUE sym = RARRAY_AREF(members, i);
+        ID id = SYM2ID(sym);
 	VALUE off = LONG2NUM(i);
 
 	if (i < N_REF_FUNC) {
 	    rb_define_method_id(nstr, id, ref_func[i], 0);
 	}
 	else {
-	    define_aref_method(nstr, ptr_members[i], off);
+            define_aref_method(nstr, sym, off);
 	}
 	define_aset_method(nstr, ID2SYM(rb_id_attrset(id)), off);
     }
@@ -437,6 +453,7 @@ rb_struct_define_under(VALUE outer, const char *name, ...)
 /*
  *  call-seq:
  *    Struct.new([class_name] [, member_name]+)                        -> StructClass
+ *    Struct.new([class_name] [, member_name]+, keyword_init: true)    -> StructClass
  *    Struct.new([class_name] [, member_name]+) {|StructClass| block } -> StructClass
  *    StructClass.new(value, ...)                                      -> object
  *    StructClass[value, ...]                                          -> object
@@ -463,6 +480,13 @@ rb_struct_define_under(VALUE outer, const char *name, ...)
  *     Customer.new("Dave", "123 Main")
  *     #=> #<struct Customer name="Dave", address="123 Main">
  *
+ *  If the optional +keyword_init+ keyword argument is set to +true+,
+ *  .new takes keyword arguments instead of normal arguments.
+ *
+ *     Customer = Struct.new(:name, :address, keyword_init: true)
+ *     Customer.new(name: "Dave", address: "123 Main")
+ *     #=> #<struct Customer name="Dave", address="123 Main">
+ *
  *  If a block is given it will be evaluated in the context of
  *  +StructClass+, passing the created class as a parameter:
  *
@@ -492,7 +516,7 @@ rb_struct_define_under(VALUE outer, const char *name, ...)
 static VALUE
 rb_struct_s_def(int argc, VALUE *argv, VALUE klass)
 {
-    VALUE name, rest;
+    VALUE name, rest, keyword_init;
     long i;
     VALUE st;
     st_table *tbl;
@@ -506,6 +530,22 @@ rb_struct_s_def(int argc, VALUE *argv, VALUE klass)
 	--argc;
 	++argv;
     }
+
+    if (RB_TYPE_P(argv[argc-1], T_HASH)) {
+	VALUE kwargs[1];
+	static ID keyword_ids[1];
+
+	if (!keyword_ids[0]) {
+	    keyword_ids[0] = rb_intern("keyword_init");
+	}
+	rb_get_kwargs(argv[argc-1], keyword_ids, 0, 1, kwargs);
+	--argc;
+	keyword_init = kwargs[0];
+    }
+    else {
+	keyword_init = Qfalse;
+    }
+
     rest = rb_ident_hash_new();
     RBASIC_CLEAR_CLASS(rest);
     tbl = RHASH_TBL(rest);
@@ -526,6 +566,7 @@ rb_struct_s_def(int argc, VALUE *argv, VALUE klass)
 	st = new_struct(name, klass);
     }
     setup_struct(st, rest);
+    rb_ivar_set(st, id_keyword_init, keyword_init);
     if (rb_block_given_p()) {
 	rb_mod_module_eval(0, 0, st);
     }
@@ -547,6 +588,31 @@ num_members(VALUE klass)
 /*
  */
 
+struct struct_hash_set_arg {
+    VALUE self;
+    VALUE unknown_keywords;
+};
+
+static int rb_struct_pos(VALUE s, VALUE *name);
+
+static int
+struct_hash_set_i(VALUE key, VALUE val, VALUE arg)
+{
+    struct struct_hash_set_arg *args = (struct struct_hash_set_arg *)arg;
+    int i = rb_struct_pos(args->self, &key);
+    if (i < 0) {
+	if (args->unknown_keywords == Qnil) {
+	    args->unknown_keywords = rb_ary_new();
+	}
+	rb_ary_push(args->unknown_keywords, key);
+    }
+    else {
+	rb_struct_modify(args->self);
+	RSTRUCT_SET(args->self, i, val);
+    }
+    return ST_CONTINUE;
+}
+
 static VALUE
 rb_struct_initialize_m(int argc, const VALUE *argv, VALUE self)
 {
@@ -555,14 +621,30 @@ rb_struct_initialize_m(int argc, const VALUE *argv, VALUE self)
 
     rb_struct_modify(self);
     n = num_members(klass);
-    if (n < argc) {
-	rb_raise(rb_eArgError, "struct size differs");
+    if (argc > 0 && RTEST(rb_struct_s_keyword_init(klass))) {
+	struct struct_hash_set_arg arg;
+	if (argc > 2 || !RB_TYPE_P(argv[0], T_HASH)) {
+	    rb_raise(rb_eArgError, "wrong number of arguments (given %d, expected 0)", argc);
+	}
+	rb_mem_clear((VALUE *)RSTRUCT_CONST_PTR(self), n);
+	arg.self = self;
+	arg.unknown_keywords = Qnil;
+	rb_hash_foreach(argv[0], struct_hash_set_i, (VALUE)&arg);
+	if (arg.unknown_keywords != Qnil) {
+	    rb_raise(rb_eArgError, "unknown keywords: %s",
+		     RSTRING_PTR(rb_ary_join(arg.unknown_keywords, rb_str_new2(", "))));
+	}
     }
-    for (i=0; i<argc; i++) {
-	RSTRUCT_SET(self, i, argv[i]);
-    }
-    if (n > argc) {
-	rb_mem_clear((VALUE *)RSTRUCT_CONST_PTR(self)+argc, n-argc);
+    else {
+	if (n < argc) {
+	    rb_raise(rb_eArgError, "struct size differs");
+	}
+	for (i=0; i<argc; i++) {
+	    RSTRUCT_SET(self, i, argv[i]);
+	}
+	if (n > argc) {
+	    rb_mem_clear((VALUE *)RSTRUCT_CONST_PTR(self)+argc, n-argc);
+	}
     }
     return Qnil;
 }
@@ -573,6 +655,43 @@ rb_struct_initialize(VALUE self, VALUE values)
     return rb_struct_initialize_m(RARRAY_LENINT(values), RARRAY_CONST_PTR(values), self);
 }
 
+static VALUE *
+struct_heap_alloc(VALUE st, size_t len)
+{
+    VALUE *ptr = rb_transient_heap_alloc((VALUE)st, sizeof(VALUE) * len);
+
+    if (ptr) {
+        RSTRUCT_TRANSIENT_SET(st);
+        return ptr;
+    }
+    else {
+        RSTRUCT_TRANSIENT_UNSET(st);
+        return ALLOC_N(VALUE, len);
+    }
+}
+
+#if USE_TRANSIENT_HEAP
+void
+rb_struct_transient_heap_evacuate(VALUE obj, int promote)
+{
+    if (RSTRUCT_TRANSIENT_P(obj)) {
+        const VALUE *old_ptr = rb_struct_const_heap_ptr(obj);
+        VALUE *new_ptr;
+        long len = RSTRUCT_LEN(obj);
+
+        if (promote) {
+            new_ptr = ALLOC_N(VALUE, len);
+            FL_UNSET_RAW(obj, RSTRUCT_TRANSIENT_FLAG);
+        }
+        else {
+            new_ptr = struct_heap_alloc(obj, len);
+        }
+        MEMCPY(new_ptr, old_ptr, VALUE, len);
+        RSTRUCT(obj)->as.heap.ptr = new_ptr;
+    }
+}
+#endif
+
 static VALUE
 struct_alloc(VALUE klass)
 {
@@ -587,9 +706,9 @@ struct_alloc(VALUE klass)
 	rb_mem_clear((VALUE *)st->as.ary, n);
     }
     else {
-	st->as.heap.ptr = ALLOC_N(VALUE, n);
-	rb_mem_clear((VALUE *)st->as.heap.ptr, n);
-	st->as.heap.len = n;
+        st->as.heap.ptr = struct_heap_alloc((VALUE)st, n);
+        rb_mem_clear((VALUE *)st->as.heap.ptr, n);
+        st->as.heap.len = n;
     }
 
     return (VALUE)st;
@@ -782,13 +901,19 @@ rb_struct_to_a(VALUE s)
 
 /*
  *  call-seq:
- *     struct.to_h     -> hash
+ *     struct.to_h                        -> hash
+ *     struct.to_h {|name, value| block } -> hash
  *
  *  Returns a Hash containing the names and values for the struct's members.
  *
+ *  If a block is given, the results of the block on each pair of the receiver
+ *  will be used as pairs.
+ *
  *     Customer = Struct.new(:name, :address, :zip)
  *     joe = Customer.new("Joe Smith", "123 Maple, Anytown NC", 12345)
  *     joe.to_h[:address]   #=> "123 Maple, Anytown NC"
+ *     joe.to_h{|name, value| [name.upcase, value.to_s.upcase]}[:ADDRESS]
+ *                          #=> "123 MAPLE, ANYTOWN NC"
  */
 
 static VALUE
@@ -797,9 +922,14 @@ rb_struct_to_h(VALUE s)
     VALUE h = rb_hash_new_with_size(RSTRUCT_LEN(s));
     VALUE members = rb_struct_members(s);
     long i;
+    int block_given = rb_block_given_p();
 
     for (i=0; i<RSTRUCT_LEN(s); i++) {
-	rb_hash_aset(h, rb_ary_entry(members, i), RSTRUCT_GET(s, i));
+        VALUE k = rb_ary_entry(members, i), v = RSTRUCT_GET(s, i);
+        if (block_given)
+            rb_hash_set_pair(h, rb_yield_values(2, k, v));
+        else
+            rb_hash_aset(h, k, v);
     }
     return h;
 }
@@ -976,6 +1106,8 @@ rb_struct_values_at(int argc, VALUE *argv, VALUE s)
  *  call-seq:
  *     struct.select {|obj| block }  -> array
  *     struct.select                 -> enumerator
+ *     struct.filter {|obj| block }  -> array
+ *     struct.filter                 -> enumerator
  *
  *  Yields each member value from the struct to the block and returns an Array
  *  containing the member values from the +struct+ for which the given block
@@ -984,6 +1116,8 @@ rb_struct_values_at(int argc, VALUE *argv, VALUE s)
  *     Lots = Struct.new(:a, :b, :c, :d, :e, :f)
  *     l = Lots.new(11, 22, 33, 44, 55, 66)
  *     l.select {|v| v.even? }   #=> [22, 44, 66]
+ *
+ *  Struct#filter is an alias for Struct#select.
  */
 
 static VALUE
@@ -1007,15 +1141,12 @@ rb_struct_select(int argc, VALUE *argv, VALUE s)
 static VALUE
 recursive_equal(VALUE s, VALUE s2, int recur)
 {
-    const VALUE *ptr, *ptr2;
     long i, len;
 
     if (recur) return Qtrue; /* Subtle! */
-    ptr = RSTRUCT_CONST_PTR(s);
-    ptr2 = RSTRUCT_CONST_PTR(s2);
     len = RSTRUCT_LEN(s);
     for (i=0; i<len; i++) {
-	if (!rb_equal(ptr[i], ptr2[i])) return Qfalse;
+        if (!rb_equal(RSTRUCT_GET(s, i), RSTRUCT_GET(s2, i))) return Qfalse;
     }
     return Qtrue;
 }
@@ -1063,31 +1194,26 @@ rb_struct_hash(VALUE s)
     long i, len;
     st_index_t h;
     VALUE n;
-    const VALUE *ptr;
 
     h = rb_hash_start(rb_hash(rb_obj_class(s)));
-    ptr = RSTRUCT_CONST_PTR(s);
     len = RSTRUCT_LEN(s);
     for (i = 0; i < len; i++) {
-	n = rb_hash(ptr[i]);
+        n = rb_hash(RSTRUCT_GET(s, i));
 	h = rb_hash_uint(h, NUM2LONG(n));
     }
     h = rb_hash_end(h);
-    return INT2FIX(h);
+    return ST2FIX(h);
 }
 
 static VALUE
 recursive_eql(VALUE s, VALUE s2, int recur)
 {
-    const VALUE *ptr, *ptr2;
     long i, len;
 
     if (recur) return Qtrue; /* Subtle! */
-    ptr = RSTRUCT_CONST_PTR(s);
-    ptr2 = RSTRUCT_CONST_PTR(s2);
     len = RSTRUCT_LEN(s);
     for (i=0; i<len; i++) {
-	if (!rb_eql(ptr[i], ptr2[i])) return Qfalse;
+        if (!rb_eql(RSTRUCT_GET(s, i), RSTRUCT_GET(s2, i))) return Qfalse;
     }
     return Qtrue;
 }
@@ -1214,10 +1340,13 @@ InitVM_Struct(void)
     rb_define_method(rb_cStruct, "[]", rb_struct_aref, 1);
     rb_define_method(rb_cStruct, "[]=", rb_struct_aset, 2);
     rb_define_method(rb_cStruct, "select", rb_struct_select, -1);
+    rb_define_method(rb_cStruct, "filter", rb_struct_select, -1);
     rb_define_method(rb_cStruct, "values_at", rb_struct_values_at, -1);
 
     rb_define_method(rb_cStruct, "members", rb_struct_members_m, 0);
     rb_define_method(rb_cStruct, "dig", rb_struct_dig, -1);
+
+    rb_define_method(rb_cStruct, "deconstruct", rb_struct_to_a, 0);
 }
 
 #undef rb_intern
@@ -1226,6 +1355,7 @@ Init_Struct(void)
 {
     id_members = rb_intern("__members__");
     id_back_members = rb_intern("__members_back__");
+    id_keyword_init = rb_intern("__keyword_init__");
 
     InitVM(Struct);
 }
