@@ -17,16 +17,36 @@ class MockClock
     def keeper_thread
       nil
     end
+
+    def stop_keeper
+      if @keeper
+        @keeper.kill
+        @keeper.join
+        @keeper = nil
+      end
+    end
   end
 
   def initialize
     @now = 2
     @reso = 1
+    @ts = nil
+    @inf = 2**31 - 1
+  end
+
+  def start_keeper
+    @now = 2
+    @reso = 1
+    @ts&.stop_keeper
     @ts = MyTS.new
     @ts.write([2, :now])
     @inf = 2**31 - 1
   end
 
+  def stop_keeper
+    @ts.stop_keeper
+  end
+
   def now
     @now.to_f
   end
@@ -100,6 +120,14 @@ class TupleSpace
 end
 
 module TupleSpaceTestModule
+  def setup
+    MockClock.instance.start_keeper
+  end
+
+  def teardown
+    MockClock.instance.stop_keeper
+  end
+
   def sleep(n)
     if Thread.current == Thread.main
       Time.forward(n)
@@ -241,17 +269,21 @@ module TupleSpaceTestModule
   end
 
   def test_ruby_talk_264062
-    th = Thread.new { @ts.take([:empty], 1) }
+    th = Thread.new {
+      assert_raise(Rinda::RequestExpiredError) do
+        @ts.take([:empty], 1)
+      end
+    }
     sleep(10)
-    assert_raise(Rinda::RequestExpiredError) do
-      thread_join(th)
-    end
+    thread_join(th)
 
-    th = Thread.new { @ts.read([:empty], 1) }
+    th = Thread.new {
+      assert_raise(Rinda::RequestExpiredError) do
+        @ts.read([:empty], 1)
+      end
+    }
     sleep(10)
-    assert_raise(Rinda::RequestExpiredError) do
-      thread_join(th)
-    end
+    thread_join(th)
   end
 
   def test_symbol_tuple
@@ -348,19 +380,18 @@ module TupleSpaceTestModule
 
     template = nil
     taker = Thread.new do
-      @ts.take([:take, nil], 10) do |t|
-        template = t
-	Thread.new do
-	  template.cancel
-	end
+      assert_raise(Rinda::RequestCanceledError) do
+        @ts.take([:take, nil], 10) do |t|
+          template = t
+          Thread.new do
+            template.cancel
+          end
+        end
       end
     end
 
     sleep(2)
-
-    assert_raise(Rinda::RequestCanceledError) do
-      assert_nil(thread_join(taker))
-    end
+    thread_join(taker)
 
     assert(template.canceled?)
 
@@ -377,19 +408,18 @@ module TupleSpaceTestModule
 
     template = nil
     reader = Thread.new do
-      @ts.read([:take, nil], 10) do |t|
-        template = t
-	Thread.new do
-	  template.cancel
-	end
+      assert_raise(Rinda::RequestCanceledError) do
+        @ts.read([:take, nil], 10) do |t|
+          template = t
+          Thread.new do
+            template.cancel
+          end
+        end
       end
     end
 
     sleep(2)
-
-    assert_raise(Rinda::RequestCanceledError) do
-      assert_nil(thread_join(reader))
-    end
+    thread_join(reader)
 
     assert(template.canceled?)
 
@@ -444,6 +474,7 @@ class TupleSpaceTest < Test::Unit::TestCase
   include TupleSpaceTestModule
 
   def setup
+    super
     ThreadGroup.new.add(Thread.current)
     @ts = Rinda::TupleSpace.new(1)
   end
@@ -455,6 +486,7 @@ class TupleSpaceTest < Test::Unit::TestCase
         th.join
       end
     }
+    super
   end
 end
 
@@ -462,6 +494,7 @@ class TupleSpaceProxyTest < Test::Unit::TestCase
   include TupleSpaceTestModule
 
   def setup
+    super
     ThreadGroup.new.add(Thread.current)
     @ts_base = Rinda::TupleSpace.new(1)
     @ts = Rinda::TupleSpaceProxy.new(@ts_base)
@@ -476,6 +509,7 @@ class TupleSpaceProxyTest < Test::Unit::TestCase
       end
     }
     @server.stop_service
+    super
   end
 
   def test_remote_array_and_hash
@@ -491,6 +525,7 @@ class TupleSpaceProxyTest < Test::Unit::TestCase
   end
 
   def test_take_bug_8215
+    skip "this test randomly fails on mswin" if /mswin/ =~ RUBY_PLATFORM
     service = DRb.start_service("druby://localhost:0", @ts_base)
 
     uri = service.uri
@@ -504,6 +539,8 @@ class TupleSpaceProxyTest < Test::Unit::TestCase
       ts = Rinda::TupleSpaceProxy.new(ro)
       th = Thread.new do
         ts.take([:test_take, nil])
+      rescue Interrupt
+        # Expected
       end
       Kernel.sleep(0.1)
       th.raise(Interrupt) # causes loss of the taken tuple
@@ -565,6 +602,8 @@ module RingIPv6
       return # IPv6 address for multicast not available
     rescue Errno::ENETDOWN
       return # Network is down
+    rescue Errno::EHOSTUNREACH
+      return # Unreachable for some reason
     end
     begin
       yield v6mc
@@ -581,6 +620,7 @@ class TestRingServer < Test::Unit::TestCase
 
     @ts = Rinda::TupleSpace.new
     @rs = Rinda::RingServer.new(@ts, [], @port)
+    @server = DRb.start_service("druby://localhost:0")
   end
   def teardown
     # implementation-dependent
@@ -591,6 +631,7 @@ class TestRingServer < Test::Unit::TestCase
       end
     }
     @rs.shutdown
+    @server.stop_service
   end
 
   def test_do_reply
@@ -616,6 +657,7 @@ class TestRingServer < Test::Unit::TestCase
   end
 
   def test_do_reply_local
+    skip 'timeout-based test becomes unstable with --jit-wait' if RubyVM::MJIT.enabled?
     with_timeout(10) {_test_do_reply_local}
   end
 
@@ -757,7 +799,10 @@ class TestRingServer < Test::Unit::TestCase
       mth.raise(Timeout::Error)
     end
     tl0 << th
+    yield
   rescue Timeout::Error => e
+    $stderr.puts "TestRingServer#with_timeout: timeout in #{n}s:"
+    $stderr.puts caller
     if tl
       bt = e.backtrace
       tl.each do |t|
@@ -842,4 +887,3 @@ class TestRingFinger < Test::Unit::TestCase
 end
 
 end
-
